@@ -7,7 +7,7 @@ extern "C" {
 #endif
 
 
-#define offsetof(TYPE,MEMBER)	((size_t)&((TYPE*)0)->MEMBER)
+#define offsetof(TYPE,MEMBER)	__builtin_offsetof(TYPE,MEMBER)
 
 #if !defined(__cplusplus)
 
