@@ -2,6 +2,7 @@
  *  npruntime.c - Scripting plugins support
  *
  *  nspluginwrapper (C) 2005-2009 Gwenole Beauchesne
+ *                  (C) 2011 David Benjamin
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -30,12 +31,6 @@
 #include "debug.h"
 
 
-// Define to enable NPClass::HasMethod cache
-#define USE_NPCLASS_HAS_METHOD_CACHE 1
-
-// Define to enable NPClass::HasProperty cache (derived from ::HasMethod cache)
-#define USE_NPCLASS_HAS_PROPERTY_CACHE 1
-
 // Defined in npw-{wrapper,viewer}.c
 extern rpc_connection_t *g_rpc_connection attribute_hidden;
 
@@ -65,26 +60,147 @@ bool npruntime_use_cache(void)
   return use_cache;
 }
 
-static inline bool use_npclass_has_method_cache(void)
+/* ====================================================================== */
+/* === NPObject stubs                                                 === */
+/* ====================================================================== */
+
+static GHashTable *g_stubs = NULL;  // uint32_t  -> (NPObjectStub *)
+
+typedef struct _NPObjectStub {
+  NPObject *npobject;
+  uint32_t id;
+} NPObjectStub;
+
+uint32_t npobject_create_stub(NPObject *npobj)
 {
-#if USE_NPCLASS_HAS_METHOD_CACHE
-  return npruntime_use_cache();
-#else
-  return false;
-#endif
+  npw_return_val_if_fail(npobj != NULL, 0);
+
+  static uint32_t next_id = 0;
+
+  // Allocate an id. Client and server get separate id spaces because
+  // of type field.
+  uint32_t id = ++next_id;
+
+  D(bug("npobject_create_stub: npobj=%p, id=0x%x\n", npobj, id));
+  NPObjectStub *stub = g_new0(NPObjectStub, 1);
+  stub->npobject = NPN_RetainObject(npobj);
+  stub->id = id;
+  g_hash_table_insert(g_stubs, GINT_TO_POINTER(stub->id), stub);
+
+  return stub->id;
 }
 
-static inline bool use_npclass_has_property_cache(void)
+static NPObjectStub *npobject_lookup_stub(uint32_t id)
 {
-  /* this depends on the NPClass::HasMethod cache */
-  return use_npclass_has_method_cache();
+  return g_hash_table_lookup(g_stubs, GINT_TO_POINTER(id));
 }
 
+NPObject *npobject_lookup_local(uint32_t id)
+{
+  NPObjectStub *stub = npobject_lookup_stub(id);
+  return stub ? stub->npobject : NULL;
+}
+
+static void npobject_destroy_stub_obj(NPObjectStub *stub)
+{
+  D(bugiI("npobject_destroy_stub: id=0x%x\n", stub->id));
+  g_hash_table_remove(g_stubs, GINT_TO_POINTER(stub->id));
+  NPN_ReleaseObject(stub->npobject);
+  g_free(stub);
+  D(bugiD("npobject_destroy_stub done\n"));
+}
+
+void npobject_destroy_stub(uint32_t id)
+{
+  NPObjectStub *stub = npobject_lookup_stub(id);
+  assert(stub != NULL);
+  npobject_destroy_stub_obj(stub);
+}
+
+/* ====================================================================== */
+/* === NPObject proxies                                               === */
+/* ====================================================================== */
+
+static GHashTable *g_proxies = NULL;  // uint32_t  -> (NPObjectProxy *)
+
+static NPClass npclass_bridge;
+
+typedef struct _NPObjectProxy {
+  NPObject parent;
+  uint32_t id;
+  bool is_valid;
+} NPObjectProxy;
+
+static NPObjectProxy *npobject_get_proxy(NPObject *npobj)
+{
+  if (npobj->_class != &npclass_bridge)
+	return NULL;
+  return (NPObjectProxy *)npobj;
+}
+
+NPObject *npobject_create_proxy(NPP instance, uint32_t id)
+{
+  D(bugiI("npobject_create_proxy: NPP=%p, id=0x%x\n", instance, id));
+
+  NPObject *object = NPN_CreateObject(instance, &npclass_bridge);
+  NPObjectProxy *proxy = npobject_get_proxy(object);
+  proxy->id = id;
+  proxy->is_valid = true;
+  // There isn't a huge need to track them by id. Any, really. But it
+  // does let us invalidate them all.
+  g_hash_table_insert(g_proxies, GINT_TO_POINTER(id), proxy);
+  D(bugiD("npobject_create_proxy done: obj=%p\n", object));
+  return object;
+}
+
+bool npobject_is_proxy(NPObject *npobj)
+{
+  return npobject_get_proxy(npobj) != NULL;
+}
+
+uint32_t npobject_get_proxy_id(NPObject *npobj)
+{
+  NPObjectProxy *proxy = npobject_get_proxy(npobj);
+  if (proxy == NULL)
+	return 0;
+  return proxy->id;
+}
+
+static void npclass_invoke_Deallocate(NPObjectProxy *proxy);
+
+void npobject_destroy_proxy(NPObject *npobj, bool release_stub)
+{
+  // Unregister the proxy.
+  D(bugiI("npobject_destroy_proxy: npobj=%p, release_stub=%d\n",
+		  npobj, release_stub));
+  NPObjectProxy *proxy = npobject_get_proxy(npobj);
+  assert(proxy != NULL);
+  if (release_stub && proxy->is_valid) {
+	npclass_invoke_Deallocate(proxy);
+  }
+  g_hash_table_remove(g_proxies, GINT_TO_POINTER(proxy->id));
+  free(npobj);
+  D(bugiD("npobject_destroy_proxy done\n"));
+}
+
+static inline bool is_valid_npobject_proxy(NPObject *npobj)
+{
+  if (npobj == NULL)
+	return false;
+  NPObjectProxy *proxy = npobject_get_proxy(npobj);
+  if (proxy == NULL)
+	return false;
+  if (!proxy->is_valid)
+	npw_printf("ERROR: NPObject proxy %p is no longer valid!\n", npobj);
+  return proxy->is_valid;
+}
 
 /* ====================================================================== */
 /* === NPClass Bridge                                                 === */
 /* ====================================================================== */
 
+static NPObject *g_NPClass_Allocate(NPP npp, NPClass *aclass);
+static void g_NPClass_Deallocate(NPObject *npobj);
 static void g_NPClass_Invalidate(NPObject *npobj);
 static bool g_NPClass_HasMethod(NPObject *npobj, NPIdentifier name);
 static bool g_NPClass_Invoke(NPObject *npobj, NPIdentifier name, const NPVariant *args, uint32_t argCount, NPVariant *result);
@@ -93,11 +209,13 @@ static bool g_NPClass_HasProperty(NPObject *npobj, NPIdentifier name);
 static bool g_NPClass_GetProperty(NPObject *npobj, NPIdentifier name, NPVariant *result);
 static bool g_NPClass_SetProperty(NPObject *npobj, NPIdentifier name, const NPVariant *value);
 static bool g_NPClass_RemoveProperty(NPObject *npobj, NPIdentifier name);
+static bool g_NPClass_Enumerate(NPObject *npobj, NPIdentifier **value, uint32_t *count);
+static bool g_NPClass_Construct(NPObject *npobj, const NPVariant *args, uint32_t argCount, NPVariant *result);
 
-NPClass npclass_bridge = {
+static NPClass npclass_bridge = {
   NPW_NP_CLASS_STRUCT_VERSION,
-  NULL,
-  NULL,
+  g_NPClass_Allocate,
+  g_NPClass_Deallocate,
   g_NPClass_Invalidate,
   g_NPClass_HasMethod,
   g_NPClass_Invoke,
@@ -105,70 +223,87 @@ NPClass npclass_bridge = {
   g_NPClass_HasProperty,
   g_NPClass_GetProperty,
   g_NPClass_SetProperty,
-  g_NPClass_RemoveProperty
+  g_NPClass_RemoveProperty,
+  g_NPClass_Enumerate,
+  g_NPClass_Construct
 };
 
 static inline bool is_valid_npobject_class(NPObject *npobj)
 {
   if (npobj == NULL || npobj->_class == NULL)
 	return false;
-  NPObjectInfo *npobj_info = npobject_info_lookup(npobj);
-  if (npobj_info == NULL)
+#if NPW_IS_PLUGIN
+  // Reject requests on invalidated objects.
+  if (!npobject_is_registered(npobj)) {
+	npw_printf("ERROR: accessed invalidated NPObject\n");
 	return false;
-  if (!npobj_info->is_valid)
-	npw_printf("ERROR: NPObject %p is no longer valid!\n", npobj);
-  return npobj_info->is_valid;
+  }
+#endif
+  return true;
 }
 
-// NPClass::Invalidate
-int npclass_handle_Invalidate(rpc_connection_t *connection)
+// NPClass::Allocate
+NPObject *g_NPClass_Allocate(NPP npp, NPClass *aclass)
 {
-  D(bug("npclass_handle_Invalidate\n"));
+  return malloc(sizeof(NPObjectProxy));
+}
 
-  NPObject *npobj;
+// NPClass::Deallocate
+static int npclass_handle_Deallocate(rpc_connection_t *connection)
+{
+  D(bug("npclass_handle_Deallocate\n"));
+
+  uint32_t id;
   int error = rpc_method_get_args(connection,
-								  RPC_TYPE_NP_OBJECT, &npobj,
+								  RPC_TYPE_UINT32, &id,
 								  RPC_TYPE_INVALID);
 
   if (error != RPC_ERROR_NO_ERROR) {
-	npw_perror("NPClass::Invalidate() get args", error);
+	npw_perror("NPClass::Deallocate() get args", error);
 	return error;
   }
 
-  if (npobj && is_valid_npobject_class(npobj) && npobj->_class->invalidate) {
-	D(bugiI("NPClass::Invalidate(npobj %p)\n", npobj));
-	npobj->_class->invalidate(npobj);
-	D(bugiD("NPClass::Invalidate done\n"));
+  D(bugiI("NPClass:Deallocate: id=0x%x\n", id));
+  NPObjectStub *stub = npobject_lookup_stub(id);
+  if (stub != NULL) {
+	npobject_destroy_stub_obj(stub);
   }
+  D(bugiD("NPClass:Deallocate done\n"));
 
   return rpc_method_send_reply(connection, RPC_TYPE_INVALID);
 }
 
-static void npclass_invoke_Invalidate(NPObject *npobj)
+static void npclass_invoke_Deallocate(NPObjectProxy *proxy)
 {
   npw_return_if_fail(rpc_method_invoke_possible(g_rpc_connection));
 
   int error = rpc_method_invoke(g_rpc_connection,
-								RPC_METHOD_NPCLASS_INVALIDATE,
-								RPC_TYPE_NP_OBJECT, npobj,
+								RPC_METHOD_NPCLASS_DEALLOCATE,
+								RPC_TYPE_UINT32, proxy->id,
 								RPC_TYPE_INVALID);
 
   if (error != RPC_ERROR_NO_ERROR) {
-	npw_perror("NPClass::Invalidate() invoke", error);
+	npw_perror("NPClass::Deallocate() invoke", error);
 	return;
   }
 
+  // FIXME: This really could be asynchronous...
   error = rpc_method_wait_for_reply(g_rpc_connection, RPC_TYPE_INVALID);
 
   if (error != RPC_ERROR_NO_ERROR) {
-	npw_perror("NPClass::Invalidate() wait for reply", error);
+	npw_perror("NPClass::Deallocate() wait for reply", error);
 	return;
   }
 }
 
+void g_NPClass_Deallocate(NPObject *npobj)
+{
+  npobject_destroy_proxy(npobj, true);
+}
+
 void g_NPClass_Invalidate(NPObject *npobj)
 {
-  if (!is_valid_npobject_class(npobj))
+  if (!is_valid_npobject_proxy(npobj))
 	return;
 
   if (!thread_check()) {
@@ -177,12 +312,18 @@ void g_NPClass_Invalidate(NPObject *npobj)
   }
 
   D(bugiI("NPClass::Invalidate(npobj %p)\n", npobj));
-  npclass_invoke_Invalidate(npobj);
+  // Just invalidate the proxy itself. There may be multiple proxies
+  // for a plugin-side NPObject. We'll invalidate them viewer-side.
+  NPObjectProxy *proxy = npobject_get_proxy(npobj);
+  proxy->is_valid = false;
+  // Release the underlying stub now, otherwise the is_valid check
+  // will refuse to do so. The stub is of no use to us now.
+  npclass_invoke_Deallocate(proxy);
   D(bugiD("NPClass::Invalidate done\n"));
 }
 
 // NPClass::HasMethod
-int npclass_handle_HasMethod(rpc_connection_t *connection)
+static int npclass_handle_HasMethod(rpc_connection_t *connection)
 {
   D(bug("npclass_handle_HasMethod\n"));
 
@@ -205,6 +346,9 @@ int npclass_handle_HasMethod(rpc_connection_t *connection)
 	D(bugiD("NPClass::HasMethod return: %d\n", ret));
   }
 
+  if (npobj)
+	NPN_ReleaseObject(npobj);
+
   return rpc_method_send_reply(connection,
 							   RPC_TYPE_UINT32, ret,
 							   RPC_TYPE_INVALID);
@@ -217,7 +361,7 @@ static bool npclass_invoke_HasMethod(NPObject *npobj, NPIdentifier name)
   int error = rpc_method_invoke(g_rpc_connection,
 								RPC_METHOD_NPCLASS_HAS_METHOD,
 								RPC_TYPE_NP_OBJECT, npobj,
-								RPC_TYPE_NP_IDENTIFIER, name,
+								RPC_TYPE_NP_IDENTIFIER, &name,
 								RPC_TYPE_INVALID);
 
   if (error != RPC_ERROR_NO_ERROR) {
@@ -236,25 +380,9 @@ static bool npclass_invoke_HasMethod(NPObject *npobj, NPIdentifier name)
   return ret;
 }
 
-static bool npclass_cached_HasMethod(NPObject *npobj, NPIdentifier name)
-{
-  NPObjectInfo *npobj_info = npobject_info_lookup(npobj);
-  if (use_npclass_has_method_cache() && npobj_info) {
-	if (G_UNLIKELY(npobj_info->hasMethod_cache == NULL))
-	  npobj_info->hasMethod_cache = g_hash_table_new(NULL, NULL);
-	gpointer hasMethod = NULL;
-	if (g_hash_table_lookup_extended(npobj_info->hasMethod_cache, name, NULL, &hasMethod))
-	  return GPOINTER_TO_UINT(hasMethod);
-  }
-  bool hasMethod = npclass_invoke_HasMethod(npobj, name);
-  if (use_npclass_has_method_cache() && npobj_info)
-	g_hash_table_insert(npobj_info->hasMethod_cache, name, GUINT_TO_POINTER(hasMethod));
-  return hasMethod;
-}
-
 bool g_NPClass_HasMethod(NPObject *npobj, NPIdentifier name)
 {
-  if (!is_valid_npobject_class(npobj))
+  if (!is_valid_npobject_proxy(npobj))
 	return false;
 
   if (!thread_check()) {
@@ -263,13 +391,13 @@ bool g_NPClass_HasMethod(NPObject *npobj, NPIdentifier name)
   }
 
   D(bugiI("NPClass::HasMethod(npobj %p, name id %p)\n", npobj, name));
-  bool ret = npclass_cached_HasMethod(npobj, name);
+  bool ret = npclass_invoke_HasMethod(npobj, name);
   D(bugiD("NPClass::HasMethod return: %d\n", ret));
   return ret;
 }
 
 // NPClass::Invoke
-int npclass_handle_Invoke(rpc_connection_t *connection)
+static int npclass_handle_Invoke(rpc_connection_t *connection)
 {
   D(bug("npclass_handle_Invoke\n"));
 
@@ -300,19 +428,18 @@ int npclass_handle_Invoke(rpc_connection_t *connection)
 	g_free(result_str);
   }
 
-  int rpc_ret = rpc_method_send_reply(connection,
-									  RPC_TYPE_UINT32, ret,
-									  RPC_TYPE_NP_VARIANT, &result,
-									  RPC_TYPE_INVALID);
-
+  if (npobj)
+	NPN_ReleaseObject(npobj);
   if (args) {
 	for (int i = 0; i < argCount; i++)
 	  NPN_ReleaseVariantValue(&args[i]);
 	free(args);
   }
 
-  NPN_ReleaseVariantValue(&result);
-  return rpc_ret;
+  return rpc_method_send_reply(connection,
+							   RPC_TYPE_UINT32, ret,
+							   RPC_TYPE_NP_VARIANT_PASS_REF, &result,
+							   RPC_TYPE_INVALID);
 }
 
 static bool npclass_invoke_Invoke(NPObject *npobj, NPIdentifier name, const NPVariant *args, uint32_t argCount,
@@ -323,7 +450,7 @@ static bool npclass_invoke_Invoke(NPObject *npobj, NPIdentifier name, const NPVa
   int error = rpc_method_invoke(g_rpc_connection,
 								RPC_METHOD_NPCLASS_INVOKE,
 								RPC_TYPE_NP_OBJECT, npobj,
-								RPC_TYPE_NP_IDENTIFIER, name,
+								RPC_TYPE_NP_IDENTIFIER, &name,
 								RPC_TYPE_ARRAY, RPC_TYPE_NP_VARIANT, argCount, args,
 								RPC_TYPE_INVALID);
 
@@ -335,7 +462,7 @@ static bool npclass_invoke_Invoke(NPObject *npobj, NPIdentifier name, const NPVa
   uint32_t ret;
   error = rpc_method_wait_for_reply(g_rpc_connection,
 									RPC_TYPE_UINT32, &ret,
-									RPC_TYPE_NP_VARIANT, result,
+									RPC_TYPE_NP_VARIANT_PASS_REF, result,
 									RPC_TYPE_INVALID);
 
   if (error != RPC_ERROR_NO_ERROR) {
@@ -353,7 +480,7 @@ bool g_NPClass_Invoke(NPObject *npobj, NPIdentifier name, const NPVariant *args,
 	return false;
   VOID_TO_NPVARIANT(*result);
 
-  if (!is_valid_npobject_class(npobj))
+  if (!is_valid_npobject_proxy(npobj))
 	return false;
 
   if (!thread_check()) {
@@ -371,7 +498,7 @@ bool g_NPClass_Invoke(NPObject *npobj, NPIdentifier name, const NPVariant *args,
 }
 
 // NPClass::InvokeDefault
-int npclass_handle_InvokeDefault(rpc_connection_t *connection)
+static int npclass_handle_InvokeDefault(rpc_connection_t *connection)
 {
   D(bug("npclass_handle_InvokeDefault\n"));
 
@@ -400,19 +527,18 @@ int npclass_handle_InvokeDefault(rpc_connection_t *connection)
 	g_free(result_str);
   }
 
-  int rpc_ret = rpc_method_send_reply(connection,
-									  RPC_TYPE_UINT32, ret,
-									  RPC_TYPE_NP_VARIANT, &result,
-									  RPC_TYPE_INVALID);
-
+  if (npobj)
+	NPN_ReleaseObject(npobj);
   if (args) {
 	for (int i = 0; i < argCount; i++)
 	  NPN_ReleaseVariantValue(&args[i]);
 	free(args);
   }
 
-  NPN_ReleaseVariantValue(&result);
-  return rpc_ret;
+  return rpc_method_send_reply(connection,
+							   RPC_TYPE_UINT32, ret,
+							   RPC_TYPE_NP_VARIANT_PASS_REF, &result,
+							   RPC_TYPE_INVALID);
 }
 
 static bool npclass_invoke_InvokeDefault(NPObject *npobj, const NPVariant *args, uint32_t argCount,
@@ -434,7 +560,7 @@ static bool npclass_invoke_InvokeDefault(NPObject *npobj, const NPVariant *args,
   uint32_t ret;
   error = rpc_method_wait_for_reply(g_rpc_connection,
 									RPC_TYPE_UINT32, &ret,
-									RPC_TYPE_NP_VARIANT, result,
+									RPC_TYPE_NP_VARIANT_PASS_REF, result,
 									RPC_TYPE_INVALID);
 
   if (error != RPC_ERROR_NO_ERROR) {
@@ -452,7 +578,7 @@ bool g_NPClass_InvokeDefault(NPObject *npobj, const NPVariant *args, uint32_t ar
 	return false;
   VOID_TO_NPVARIANT(*result);
 
-  if (!is_valid_npobject_class(npobj))
+  if (!is_valid_npobject_proxy(npobj))
 	return false;
 
   if (!thread_check()) {
@@ -470,7 +596,7 @@ bool g_NPClass_InvokeDefault(NPObject *npobj, const NPVariant *args, uint32_t ar
 }
 
 // NPClass::HasProperty
-int npclass_handle_HasProperty(rpc_connection_t *connection)
+static int npclass_handle_HasProperty(rpc_connection_t *connection)
 {
   D(bug("npclass_handle_HasProperty\n"));
 
@@ -493,6 +619,9 @@ int npclass_handle_HasProperty(rpc_connection_t *connection)
 	D(bugiD("NPClass::HasProperty return: %d\n", ret));
   }
 
+  if (npobj)
+	NPN_ReleaseObject(npobj);
+
   return rpc_method_send_reply(connection,
 							   RPC_TYPE_UINT32, ret,
 							   RPC_TYPE_INVALID);
@@ -505,7 +634,7 @@ static bool npclass_invoke_HasProperty(NPObject *npobj, NPIdentifier name)
   int error = rpc_method_invoke(g_rpc_connection,
 								RPC_METHOD_NPCLASS_HAS_PROPERTY,
 								RPC_TYPE_NP_OBJECT, npobj,
-								RPC_TYPE_NP_IDENTIFIER, name,
+								RPC_TYPE_NP_IDENTIFIER, &name,
 								RPC_TYPE_INVALID);
 
   if (error != RPC_ERROR_NO_ERROR) {
@@ -524,20 +653,9 @@ static bool npclass_invoke_HasProperty(NPObject *npobj, NPIdentifier name)
   return ret;
 }
 
-static bool npclass_cached_HasProperty(NPObject *npobj, NPIdentifier name)
-{
-  NPObjectInfo *npobj_info = npobject_info_lookup(npobj);
-  if (use_npclass_has_property_cache() && npobj_info && npobj_info->hasMethod_cache) {
-	/* If the NPIdentifier references a method, it can't be a property */
-	if (g_hash_table_lookup_extended(npobj_info->hasMethod_cache, name, NULL, NULL))
-	  return false;
-  }
-  return npclass_invoke_HasProperty(npobj, name);
-}
-
 bool g_NPClass_HasProperty(NPObject *npobj, NPIdentifier name)
 {
-  if (!is_valid_npobject_class(npobj))
+  if (!is_valid_npobject_proxy(npobj))
 	return false;
 
   if (!thread_check()) {
@@ -546,13 +664,13 @@ bool g_NPClass_HasProperty(NPObject *npobj, NPIdentifier name)
   }
 
   D(bugiI("NPClass::HasProperty(npobj %p, name id %p)\n", npobj, name));
-  bool ret = npclass_cached_HasProperty(npobj, name);
+  bool ret = npclass_invoke_HasProperty(npobj, name);
   D(bugiD("NPClass::HasProperty return: %d\n", ret));
   return ret;
 }
   
 // NPClass::GetProperty
-int npclass_handle_GetProperty(rpc_connection_t *connection)
+static int npclass_handle_GetProperty(rpc_connection_t *connection)
 {
   D(bug("npclass_handle_GetProperty\n"));
 
@@ -579,13 +697,13 @@ int npclass_handle_GetProperty(rpc_connection_t *connection)
 	g_free(result_str);
   }
 
-  int rpc_ret = rpc_method_send_reply(connection,
-									  RPC_TYPE_UINT32, ret,
-									  RPC_TYPE_NP_VARIANT, &result,
-									  RPC_TYPE_INVALID);
+  if (npobj)
+	NPN_ReleaseObject(npobj);
 
-  NPN_ReleaseVariantValue(&result);
-  return rpc_ret;
+  return rpc_method_send_reply(connection,
+							   RPC_TYPE_UINT32, ret,
+							   RPC_TYPE_NP_VARIANT_PASS_REF, &result,
+							   RPC_TYPE_INVALID);
 }
 
 static bool npclass_invoke_GetProperty(NPObject *npobj, NPIdentifier name, NPVariant *result)
@@ -595,7 +713,7 @@ static bool npclass_invoke_GetProperty(NPObject *npobj, NPIdentifier name, NPVar
   int error = rpc_method_invoke(g_rpc_connection,
 								RPC_METHOD_NPCLASS_GET_PROPERTY,
 								RPC_TYPE_NP_OBJECT, npobj,
-								RPC_TYPE_NP_IDENTIFIER, name,
+								RPC_TYPE_NP_IDENTIFIER, &name,
 								RPC_TYPE_INVALID);
 
   if (error != RPC_ERROR_NO_ERROR) {
@@ -606,7 +724,7 @@ static bool npclass_invoke_GetProperty(NPObject *npobj, NPIdentifier name, NPVar
   uint32_t ret;
   error = rpc_method_wait_for_reply(g_rpc_connection,
 									RPC_TYPE_UINT32, &ret,
-									RPC_TYPE_NP_VARIANT, result,
+									RPC_TYPE_NP_VARIANT_PASS_REF, result,
 									RPC_TYPE_INVALID);
 
   if (error != RPC_ERROR_NO_ERROR) {
@@ -623,7 +741,7 @@ bool g_NPClass_GetProperty(NPObject *npobj, NPIdentifier name, NPVariant *result
 	return false;
   VOID_TO_NPVARIANT(*result);
 
-  if (!is_valid_npobject_class(npobj))
+  if (!is_valid_npobject_proxy(npobj))
 	return false;
 
   if (!thread_check()) {
@@ -640,7 +758,7 @@ bool g_NPClass_GetProperty(NPObject *npobj, NPIdentifier name, NPVariant *result
 }
   
 // NPClass::SetProperty
-int npclass_handle_SetProperty(rpc_connection_t *connection)
+static int npclass_handle_SetProperty(rpc_connection_t *connection)
 {
   D(bug("npclass_handle_SetProperty\n"));
 
@@ -665,12 +783,13 @@ int npclass_handle_SetProperty(rpc_connection_t *connection)
 	D(bugiD("NPClass::SetProperty return: %d\n", ret));
   }
 
-  int rpc_ret = rpc_method_send_reply(connection,
-									  RPC_TYPE_UINT32, ret,
-									  RPC_TYPE_INVALID);
-
+  if (npobj)
+	NPN_ReleaseObject(npobj);
   NPN_ReleaseVariantValue(&value);
-  return rpc_ret;
+
+  return rpc_method_send_reply(connection,
+							   RPC_TYPE_UINT32, ret,
+							   RPC_TYPE_INVALID);
 }
 
 static bool npclass_invoke_SetProperty(NPObject *npobj, NPIdentifier name, const NPVariant *value)
@@ -680,7 +799,7 @@ static bool npclass_invoke_SetProperty(NPObject *npobj, NPIdentifier name, const
   int error = rpc_method_invoke(g_rpc_connection,
 								RPC_METHOD_NPCLASS_SET_PROPERTY,
 								RPC_TYPE_NP_OBJECT, npobj,
-								RPC_TYPE_NP_IDENTIFIER, name,
+								RPC_TYPE_NP_IDENTIFIER, &name,
 								RPC_TYPE_NP_VARIANT, value,
 								RPC_TYPE_INVALID);
 
@@ -709,7 +828,7 @@ bool g_NPClass_SetProperty(NPObject *npobj, NPIdentifier name, const NPVariant *
 	return false;
   }
 
-  if (!is_valid_npobject_class(npobj))
+  if (!is_valid_npobject_proxy(npobj))
 	return false;
 
   if (!thread_check()) {
@@ -724,7 +843,7 @@ bool g_NPClass_SetProperty(NPObject *npobj, NPIdentifier name, const NPVariant *
 }
 
 // NPClass::RemoveProperty
-int npclass_handle_RemoveProperty(rpc_connection_t *connection)
+static int npclass_handle_RemoveProperty(rpc_connection_t *connection)
 {
   D(bug("npclass_handle_RemoveProperty\n"));
 
@@ -747,6 +866,9 @@ int npclass_handle_RemoveProperty(rpc_connection_t *connection)
 	D(bugiD("NPClass::RemoveProperty return: %d\n", ret));
   }
 
+  if (npobj)
+	NPN_ReleaseObject(npobj);
+
   return rpc_method_send_reply(connection,
 							   RPC_TYPE_UINT32, ret,
 							   RPC_TYPE_INVALID);
@@ -759,7 +881,7 @@ static bool npclass_invoke_RemoveProperty(NPObject *npobj, NPIdentifier name)
   int error = rpc_method_invoke(g_rpc_connection,
 								RPC_METHOD_NPCLASS_REMOVE_PROPERTY,
 								RPC_TYPE_NP_OBJECT, npobj,
-								RPC_TYPE_NP_IDENTIFIER, name,
+								RPC_TYPE_NP_IDENTIFIER, &name,
 								RPC_TYPE_INVALID);
 
   if (error != RPC_ERROR_NO_ERROR) {
@@ -780,7 +902,7 @@ static bool npclass_invoke_RemoveProperty(NPObject *npobj, NPIdentifier name)
 
 bool g_NPClass_RemoveProperty(NPObject *npobj, NPIdentifier name)
 {
-  if (!is_valid_npobject_class(npobj))
+  if (!is_valid_npobject_proxy(npobj))
 	return false;
 
   if (!thread_check()) {
@@ -794,171 +916,307 @@ bool g_NPClass_RemoveProperty(NPObject *npobj, NPIdentifier name)
   return ret;
 }
 
-
-/* ====================================================================== */
-/* === NPObjectInfo                                                   === */
-/* ====================================================================== */
-
-NPObjectInfo *npobject_info_new(NPObject *npobj)
+// NPClass::Enumerate
+static int npclass_handle_Enumerate(rpc_connection_t *connection)
 {
-  NPObjectInfo *npobj_info = NPW_MemNew(NPObjectInfo, 1);
-  if (npobj_info) {
-	static uint32_t id;
-	npobj_info->npobj = npobj;
-	npobj_info->npobj_id = ++id;
-	npobj_info->is_valid = true;
-	npobj_info->plugin = NULL;
-	npobj_info->hasMethod_cache = NULL;
-  }
-  return npobj_info;
-}
+  D(bug("npclass_handle_Enumerate\n"));
 
-void npobject_info_destroy(NPObjectInfo *npobj_info)
-{
-  if (npobj_info == NULL)
-	return;
-
-  npw_plugin_instance_unref(npobj_info->plugin);
-
-  if (npobj_info->hasMethod_cache) {
-	g_hash_table_destroy(npobj_info->hasMethod_cache);
-	npobj_info->hasMethod_cache = NULL;
-  }
-
-  NPW_MemFree(npobj_info);
-}
-
-
-/* ====================================================================== */
-/* === NPObject                                                       === */
-/* ====================================================================== */
-
-static void npobject_hash_table_insert(NPObject *npobj, NPObjectInfo *npobj_info);
-static bool npobject_hash_table_remove(NPObject *npobj);
-
-static NPObject *_npobject_new(NPP instance, NPClass *class)
-{
   NPObject *npobj;
-  if (class && class->allocate)
-	npobj = class->allocate(instance, class);
-  else
-	npobj = malloc(sizeof(*npobj));
-  if (npobj) {
-	npobj->_class = class ? class : &npclass_bridge;
-	npobj->referenceCount = 1;
+  int error = rpc_method_get_args(connection,
+								  RPC_TYPE_NP_OBJECT, &npobj,
+								  RPC_TYPE_INVALID);
+
+  if (error != RPC_ERROR_NO_ERROR) {
+	npw_perror("NPClass::Enumerate() get args", error);
+	return error;
   }
-  return npobj;
-}
 
-static void _npobject_destroy(NPObject *npobj)
-{
-  if (npobj) {
-	if (npobj->_class && npobj->_class->deallocate)
-	  npobj->_class->deallocate(npobj);
-	else
-	  free(npobj);
+  uint32_t ret = false;
+  NPIdentifier *idents = NULL;
+  uint32_t argCount = 0;
+  if (npobj && is_valid_npobject_class(npobj) &&
+	  NP_CLASS_STRUCT_VERSION_HAS_ENUM(npobj->_class) &&
+	  npobj->_class->enumerate) {
+	D(bugiI("NPClass::Enumerate(npobj %p)\n", npobj));
+	ret = npobj->_class->enumerate(npobj, &idents, &argCount);
+	// TODO: Print the identifiers to debug
+	D(bugiD("NPClass::Enumerate return: %d (%d identifiers)\n", ret, argCount));
   }
-}
 
-NPObject *npobject_new(uint32_t npobj_id, NPP instance, NPClass *class)
-{
-  NPObject *npobj = _npobject_new(instance, class);
-  if (npobj == NULL)
-	return NULL;
-
-  NPObjectInfo *npobj_info = npobject_info_new(npobj);
-  if (npobj_info == NULL) {
-	_npobject_destroy(npobj);
-	return NULL;
-  }
-  npobj_info->npobj_id = npobj_id;
-  npobj_info->plugin = npw_plugin_instance_ref(NPW_PLUGIN_INSTANCE(instance));
-  npobject_associate(npobj, npobj_info);
-  return npobj;
-}
-
-void npobject_destroy(NPObject *npobj)
-{
   if (npobj)
-	npobject_hash_table_remove(npobj);
+	NPN_ReleaseObject(npobj);
 
-  _npobject_destroy(npobj);
+  int rpc_ret = rpc_method_send_reply(connection,
+									  RPC_TYPE_UINT32, ret,
+									  RPC_TYPE_ARRAY, RPC_TYPE_NP_IDENTIFIER, argCount, idents,
+									  RPC_TYPE_INVALID);
+  if (idents) NPN_MemFree(idents);
+
+  return rpc_ret;
 }
 
-void npobject_associate(NPObject *npobj, NPObjectInfo *npobj_info)
+static bool npclass_invoke_Enumerate(NPObject *npobj,
+									 NPIdentifier **idents, uint32_t *count)
 {
-  assert(npobj && npobj_info && npobj_info->npobj_id > 0);
-  npobject_hash_table_insert(npobj, npobj_info);
+  npw_return_val_if_fail(rpc_method_invoke_possible(g_rpc_connection), false);
+
+  int error = rpc_method_invoke(g_rpc_connection,
+								RPC_METHOD_NPCLASS_ENUMERATE,
+								RPC_TYPE_NP_OBJECT, npobj,
+								RPC_TYPE_INVALID);
+
+  if (error != RPC_ERROR_NO_ERROR) {
+	npw_perror("NPClass::Enumerate() invoke", error);
+	return false;
+  }
+
+  uint32_t ret;
+  uint32_t myCount = 0;
+  NPIdentifier *myIdentifiers = NULL;
+  error = rpc_method_wait_for_reply(g_rpc_connection,
+									RPC_TYPE_UINT32, &ret,
+									RPC_TYPE_ARRAY, RPC_TYPE_NP_IDENTIFIER, &myCount, &myIdentifiers,
+									RPC_TYPE_INVALID);
+
+  if (error != RPC_ERROR_NO_ERROR) {
+	npw_perror("NPClass::Enumerate() wait for reply", error);
+	return false;
+  }
+
+  *count = myCount;
+  if (ret) {
+	ret = NPW_ReallocData(myIdentifiers,
+						  sizeof(**idents) * myCount,
+						  (void**)idents) == NPERR_NO_ERROR;
+  }
+  if (myIdentifiers)
+	free(myIdentifiers);
+
+  return ret;
 }
 
+bool g_NPClass_Enumerate(NPObject *npobj,
+						 NPIdentifier **idents, uint32_t *count)
+{
+  if (count == NULL || idents == NULL)
+	return false;
 
+  if (!is_valid_npobject_proxy(npobj))
+	return false;
+
+  if (!thread_check()) {
+	npw_printf("WARNING: NPClass::Enumerate not called from the main thread\n");
+	return false;
+  }
+
+  D(bugiI("NPClass::Enumerate(npobj %p)\n", npobj));
+  bool ret = npclass_invoke_Enumerate(npobj, idents, count);
+  // TODO: print the identifiers to debug
+  D(bugiD("NPClass::Enumerate return: %d (%d)\n", ret, *count));
+  return ret;
+}
+
+// NPClass::Construct
+static int npclass_handle_Construct(rpc_connection_t *connection)
+{
+  D(bug("npclass_handle_Construct\n"));
+
+  NPObject *npobj;
+  uint32_t argCount;
+  NPVariant *args;
+  int error = rpc_method_get_args(connection,
+								  RPC_TYPE_NP_OBJECT, &npobj,
+								  RPC_TYPE_ARRAY, RPC_TYPE_NP_VARIANT, &argCount, &args,
+								  RPC_TYPE_INVALID);
+
+  if (error != RPC_ERROR_NO_ERROR) {
+	npw_perror("NPClass::Construct() get args", error);
+	return error;
+  }
+
+  uint32_t ret = false;
+  NPVariant result;
+  VOID_TO_NPVARIANT(result);
+  if (npobj && is_valid_npobject_class(npobj) && npobj->_class->construct) {
+	D(bugiI("NPClass::Construct(npobj %p)\n", npobj));
+	print_npvariant_args(args, argCount);
+	ret = npobj->_class->construct(npobj, args, argCount, &result);
+	gchar *result_str = string_of_NPVariant(&result);
+	D(bugiD("NPClass::Construct return: %d (%s)\n", ret, result_str));
+	g_free(result_str);
+  }
+
+  if (npobj)
+	NPN_ReleaseObject(npobj);
+  if (args) {
+	for (int i = 0; i < argCount; i++)
+	  NPN_ReleaseVariantValue(&args[i]);
+	free(args);
+  }
+
+  return rpc_method_send_reply(connection,
+							   RPC_TYPE_UINT32, ret,
+							   RPC_TYPE_NP_VARIANT_PASS_REF, &result,
+							   RPC_TYPE_INVALID);
+}
+
+static bool npclass_invoke_Construct(NPObject *npobj, const NPVariant *args, uint32_t argCount,
+										 NPVariant *result)
+{
+  npw_return_val_if_fail(rpc_method_invoke_possible(g_rpc_connection), false);
+
+  int error = rpc_method_invoke(g_rpc_connection,
+								RPC_METHOD_NPCLASS_CONSTRUCT,
+								RPC_TYPE_NP_OBJECT, npobj,
+								RPC_TYPE_ARRAY, RPC_TYPE_NP_VARIANT, argCount, args,
+								RPC_TYPE_INVALID);
+
+  if (error != RPC_ERROR_NO_ERROR) {
+	npw_perror("NPClass::Construct() invoke", error);
+	return false;
+  }
+
+  uint32_t ret;
+  error = rpc_method_wait_for_reply(g_rpc_connection,
+									RPC_TYPE_UINT32, &ret,
+									RPC_TYPE_NP_VARIANT_PASS_REF, result,
+									RPC_TYPE_INVALID);
+
+  if (error != RPC_ERROR_NO_ERROR) {
+	npw_perror("NPClass::Construct() wait for reply", error);
+	return false;
+  }
+
+  return ret;
+}
+
+bool g_NPClass_Construct(NPObject *npobj, const NPVariant *args, uint32_t argCount,
+						 NPVariant *result)
+{
+  if (result == NULL)
+	return false;
+  VOID_TO_NPVARIANT(*result);
+
+  if (!is_valid_npobject_proxy(npobj))
+	return false;
+
+  if (!thread_check()) {
+	npw_printf("WARNING: NPClass::Construct not called from the main thread\n");
+	return false;
+  }
+
+  D(bugiI("NPClass::Construct(npobj %p)\n", npobj));
+  print_npvariant_args(args, argCount);
+  bool ret = npclass_invoke_Construct(npobj, args, argCount, result);
+  gchar *result_str = string_of_NPVariant(result);
+  D(bugiD("NPClass::Construct return: %d (%s)\n", ret, result_str));
+  g_free(result_str);
+  return ret;
+}
+
+int npclass_add_method_descriptors(rpc_connection_t *connection)
+{
+  static const rpc_method_descriptor_t vtable[] = {
+	{ RPC_METHOD_NPCLASS_HAS_METHOD,		npclass_handle_HasMethod },
+	{ RPC_METHOD_NPCLASS_INVOKE,			npclass_handle_Invoke },
+	{ RPC_METHOD_NPCLASS_INVOKE_DEFAULT,	npclass_handle_InvokeDefault },
+	{ RPC_METHOD_NPCLASS_HAS_PROPERTY,		npclass_handle_HasProperty },
+	{ RPC_METHOD_NPCLASS_GET_PROPERTY,		npclass_handle_GetProperty },
+	{ RPC_METHOD_NPCLASS_SET_PROPERTY,		npclass_handle_SetProperty },
+	{ RPC_METHOD_NPCLASS_REMOVE_PROPERTY,	npclass_handle_RemoveProperty },
+	{ RPC_METHOD_NPCLASS_ENUMERATE,			npclass_handle_Enumerate },
+	{ RPC_METHOD_NPCLASS_CONSTRUCT,			npclass_handle_Construct },
+	{ RPC_METHOD_NPCLASS_DEALLOCATE,		npclass_handle_Deallocate },
+  };
+  return rpc_connection_add_method_descriptors(g_rpc_connection, vtable,
+											   sizeof(vtable) / sizeof(vtable[0]));
+}
+
+#if NPW_IS_PLUGIN
 /* ====================================================================== */
-/* === NPObject Repository                                            === */
+/* === NPObject Registration                                          === */
 /* ====================================================================== */
 
-// NOTE: those hashes must be maintained in a whole, not separately
-static GHashTable *g_npobjects = NULL;			// (NPObject *)  -> (NPObjectInfo *)
-static GHashTable *g_npobject_ids = NULL;		// (NPObject ID) -> (NPObject *)
+static GHashTable *g_npobj_owners = NULL;
+
+void npobject_register(NPObject *npobj, void *plugin)
+{
+  assert(!npobject_is_registered(npobj));
+  g_hash_table_insert(g_npobj_owners, npobj, plugin);
+}
+
+bool npobject_is_registered(NPObject *npobj)
+{
+  return g_hash_table_lookup_extended(g_npobj_owners, npobj, NULL, NULL);
+}
+
+void *npobject_get_owner(NPObject *npobj)
+{
+  return g_hash_table_lookup(g_npobj_owners, npobj);
+}
+
+void npobject_unregister(NPObject *npobj)
+{
+  g_hash_table_remove(g_npobj_owners, npobj);
+}
+#endif
+
+/* ====================================================================== */
+/* === NPObject Bridge System                                         === */
+/* ====================================================================== */
 
 bool npobject_bridge_new(void)
 {
-  if ((g_npobjects = g_hash_table_new_full(NULL, NULL, NULL, (GDestroyNotify)npobject_info_destroy)) == NULL)
-	return false;
-  if ((g_npobject_ids = g_hash_table_new(NULL, NULL)) == NULL)
-	return false;
+  g_stubs = g_hash_table_new(NULL, NULL);
+  g_proxies = g_hash_table_new(NULL, NULL);
+#if NPW_IS_PLUGIN
+  g_npobj_owners = g_hash_table_new(NULL, NULL);
+#endif
   return true;
 }
 
 void npobject_bridge_destroy(void)
 {
-  if (g_npobject_ids) {
-	g_hash_table_destroy(g_npobject_ids);
-	g_npobject_ids = NULL;
+  if (g_stubs) {
+	g_hash_table_destroy(g_stubs);
+	g_stubs = NULL;
   }
-  if (g_npobjects) {
-	g_hash_table_destroy(g_npobjects);
-	g_npobjects = NULL;
+  if (g_proxies) {
+	g_hash_table_destroy(g_proxies);
+	g_proxies = NULL;
   }
+#if NPW_IS_PLUGIN
+  if (g_npobj_owners) {
+	g_hash_table_destroy(g_npobj_owners);
+	g_npobj_owners = NULL;
+  }
+#endif
 }
 
-void npobject_hash_table_insert(NPObject *npobj, NPObjectInfo *npobj_info)
+#if NPW_IS_BROWSER
+static void proxy_deactivate_func(gpointer key, gpointer value, gpointer user_data)
 {
-  g_hash_table_insert(g_npobjects, npobj, npobj_info);
-  g_hash_table_insert(g_npobject_ids, (void *)(uintptr_t)npobj_info->npobj_id, npobj);
+  NPObjectProxy *proxy = (NPObjectProxy *)value;
+  proxy->is_valid = false;
+  // No need to release the stub. At this point the viewer is dead.
 }
 
-bool npobject_hash_table_remove(NPObject *npobj)
+static void stub_destroy_func(gpointer key, gpointer value, gpointer user_data)
 {
-  NPObjectInfo *npobj_info = npobject_info_lookup(npobj);
-  assert(npobj_info != NULL);
-  bool removed_all = true;
-  if (!g_hash_table_remove(g_npobject_ids, (void *)(uintptr_t)npobj_info->npobj_id))
-	removed_all = false;
-  if (!g_hash_table_remove(g_npobjects, npobj))
-	removed_all = false;
-  return removed_all;
-}
-
-NPObjectInfo *npobject_info_lookup(NPObject *npobj)
-{
-  return g_hash_table_lookup(g_npobjects, npobj);
-}
-
-NPObject *npobject_lookup(uint32_t npobj_id)
-{
-  return g_hash_table_lookup(g_npobject_ids, (void *)(uintptr_t)npobj_id);
-}
-
-static void npruntime_deactivate_func(gpointer key, gpointer value, gpointer user_data)
-{
-  NPObjectInfo *npobj_info = (NPObjectInfo *)value;
-  npobj_info->is_valid = false;
+  NPObjectStub *stub = (NPObjectStub *)value;
+  npobject_destroy_stub_obj(stub);
 }
 
 void npruntime_deactivate(void)
 {
-  g_hash_table_foreach(g_npobjects, npruntime_deactivate_func, NULL);
+  // Invalidate any proxies the wrapper may still be holding on to.
+  g_hash_table_foreach(g_proxies, proxy_deactivate_func, NULL);
+  g_hash_table_foreach(g_stubs, stub_destroy_func, NULL);
+  // Reset both tables.
+  npobject_bridge_destroy();
+  npobject_bridge_new();
 }
+#endif
 
 
 /* ====================================================================== */
@@ -978,8 +1236,8 @@ npvariant_clear(NPVariant *variant)
   case NPVariantType_String:
 	{
 	  NPString *s = &NPVARIANT_TO_STRING(*variant);
-	  if (s->utf8characters)
-		NPN_MemFree((void *)s->utf8characters);
+	  if (s->UTF8Characters)
+		NPN_MemFree((void *)s->UTF8Characters);
 	  break;
 	}
   case NPVariantType_Object:
@@ -1021,8 +1279,8 @@ string_of_NPVariant(const NPVariant *arg)
 	case NPVariantType_String:
 	  g_string_append_c(str, '\'');
 	  g_string_append_len(str,
-						  arg->value.stringValue.utf8characters,
-						  arg->value.stringValue.utf8length);
+						  arg->value.stringValue.UTF8Characters,
+						  arg->value.stringValue.UTF8Length);
 	  g_string_append_c(str, '\'');
 	  break;
 	case NPVariantType_Object:
