@@ -50,51 +50,66 @@ enum {
   RPC_METHOD_NPN_FORCE_REDRAW,					/* 20 */
   RPC_METHOD_NPN_INVALIDATE_RECT,
   RPC_METHOD_NPN_INVALIDATE_REGION,
+  RPC_METHOD_NPN_GET_VALUE_FOR_URL,
+  RPC_METHOD_NPN_SET_VALUE_FOR_URL,
+  RPC_METHOD_NPN_GET_AUTHENTICATION_INFO,		/* 25 */
+  RPC_METHOD_NPN_POP_UP_CONTEXT_MENU,
+  RPC_METHOD_NPN_HANDLE_EVENT,
+  RPC_METHOD_NPN_UNFOCUS_INSTANCE,
+  RPC_METHOD_NPN_URL_REDIRECT_RESPONSE,
 
-  RPC_METHOD_NPP_NEW,
+  RPC_METHOD_NPP_NEW,							/* 30 */
   RPC_METHOD_NPP_DESTROY,
-  RPC_METHOD_NPP_SET_WINDOW,					/* 25 */
+  RPC_METHOD_NPP_SET_WINDOW,
   RPC_METHOD_NPP_HANDLE_EVENT,
   RPC_METHOD_NPP_GET_VALUE,
-  RPC_METHOD_NPP_SET_VALUE,
+  RPC_METHOD_NPP_SET_VALUE,						/* 35 */
   RPC_METHOD_NPP_URL_NOTIFY,
-  RPC_METHOD_NPP_NEW_STREAM,					/* 30 */
+  RPC_METHOD_NPP_NEW_STREAM,
   RPC_METHOD_NPP_DESTROY_STREAM,
   RPC_METHOD_NPP_WRITE_READY,
-  RPC_METHOD_NPP_WRITE,
+  RPC_METHOD_NPP_WRITE,							/* 40 */
   RPC_METHOD_NPP_STREAM_AS_FILE,
-  RPC_METHOD_NPP_PRINT,							/* 35 */
+  RPC_METHOD_NPP_PRINT,
+  RPC_METHOD_NPP_GOT_FOCUS,
+  RPC_METHOD_NPP_LOST_FOCUS,					/* 45 */
+  RPC_METHOD_NPP_URL_REDIRECT_NOTIFY,
+  RPC_METHOD_NPP_CLEAR_SITE_DATA,
+  RPC_METHOD_NPP_GET_SITES_WITH_DATA,
 
   RPC_METHOD_NPN_CREATE_OBJECT,
-  RPC_METHOD_NPN_RETAIN_OBJECT,
+  RPC_METHOD_NPN_RETAIN_OBJECT,					/* 50 */
   RPC_METHOD_NPN_RELEASE_OBJECT,
   RPC_METHOD_NPN_INVOKE,
-  RPC_METHOD_NPN_INVOKE_DEFAULT,				/* 40 */
+  RPC_METHOD_NPN_INVOKE_DEFAULT,
   RPC_METHOD_NPN_EVALUATE,
-  RPC_METHOD_NPN_GET_PROPERTY,
+  RPC_METHOD_NPN_GET_PROPERTY,					/* 55 */
   RPC_METHOD_NPN_SET_PROPERTY,
   RPC_METHOD_NPN_REMOVE_PROPERTY,
-  RPC_METHOD_NPN_HAS_PROPERTY,					/* 45 */
+  RPC_METHOD_NPN_HAS_PROPERTY,
   RPC_METHOD_NPN_HAS_METHOD,
-  RPC_METHOD_NPN_SET_EXCEPTION,
+  RPC_METHOD_NPN_SET_EXCEPTION,					/* 60 */
   RPC_METHOD_NPN_ENUMERATE,
   RPC_METHOD_NPN_CONSTRUCT,
 
-  RPC_METHOD_NPN_GET_STRING_IDENTIFIER,			/* 50 */
+  RPC_METHOD_NPN_GET_STRING_IDENTIFIER,
   RPC_METHOD_NPN_GET_STRING_IDENTIFIERS,
-  RPC_METHOD_NPN_GET_INT_IDENTIFIER,
+  RPC_METHOD_NPN_GET_INT_IDENTIFIER,			/* 65 */
   RPC_METHOD_NPN_IDENTIFIER_IS_STRING,
   RPC_METHOD_NPN_UTF8_FROM_IDENTIFIER,
-  RPC_METHOD_NPN_INT_FROM_IDENTIFIER,			/* 55 */
+  RPC_METHOD_NPN_INT_FROM_IDENTIFIER,
 
   RPC_METHOD_NPCLASS_INVALIDATE,
-  RPC_METHOD_NPCLASS_HAS_METHOD,
+  RPC_METHOD_NPCLASS_HAS_METHOD,				/* 70 */
   RPC_METHOD_NPCLASS_INVOKE,
   RPC_METHOD_NPCLASS_INVOKE_DEFAULT,
-  RPC_METHOD_NPCLASS_HAS_PROPERTY,				/* 60 */
+  RPC_METHOD_NPCLASS_HAS_PROPERTY,
   RPC_METHOD_NPCLASS_GET_PROPERTY,
-  RPC_METHOD_NPCLASS_SET_PROPERTY,
-  RPC_METHOD_NPCLASS_REMOVE_PROPERTY
+  RPC_METHOD_NPCLASS_SET_PROPERTY,				/* 75 */
+  RPC_METHOD_NPCLASS_REMOVE_PROPERTY,
+  RPC_METHOD_NPCLASS_ENUMERATE,
+  RPC_METHOD_NPCLASS_CONSTRUCT,
+  RPC_METHOD_NPCLASS_DEALLOCATE
 };
 
 // NPAPI data types
@@ -116,6 +131,8 @@ enum {
   RPC_TYPE_NP_STRING,							/* 15 */
   RPC_TYPE_NP_VARIANT,
   RPC_TYPE_NP_UTF8,
+  RPC_TYPE_NP_OBJECT_PASS_REF,
+  RPC_TYPE_NP_VARIANT_PASS_REF,
   RPC_TYPE_NPW_PLUGIN_INSTANCE
 };
 
